// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021-2022 Florian Fischer, Florian Schmaus
#include <fcntl.h>
#include <unistd.h>

#include <array>
#include <boost/program_options.hpp>
#include <cstdlib>
#include <cstring>
#include <exception>
#include <filesystem>
#include <iostream>
#include <sstream>
#include <string>

#include "Common.hpp"
#include "Fiber.hpp"
#include "Runtime.hpp"
#include "Semaphore.hpp"
#include "StealingMode.hpp"
#include "io.hpp"
#include "lib/ShardedFileBuffer.hpp"

namespace fs = std::filesystem;
namespace po = boost::program_options;

#define EMPER_RIPGREP_BUFSIZE 4096

static const char* needle;
static size_t needle_len;

static emper::Semaphore* max_fibers;
static emper::Semaphore* max_searching;
static emper::Semaphore* max_walking;

static emper::lib::ShardedFileBuffer* outBuf;

static enum emper::StealingMode stealingMode;

void search(const std::string& path) {
	int fd = emper::io::openAndWait(path.c_str(), O_RDONLY);
	if (fd < 0) {
		DIE_MSG_ERRNO("open of " << path << " failed");
	}

	std::array<char, EMPER_RIPGREP_BUFSIZE> buf;

	ssize_t bytes_read = emper::io::readFileAndWait(fd, buf.data(), buf.size(), 0);
	while (bytes_read > 0) {
		if (memmem(&buf[0], bytes_read, needle, needle_len)) {
			outBuf->getStream() << path << std::endl;
			goto out;
		}

		bytes_read = emper::io::readFileAndWait(fd, buf.data(), buf.size(), -1);
	}

	if (bytes_read < 0) {
		DIE_MSG_ERRNO("read of " << path << " failed");
	}

out:
	emper::io::closeAndForget(fd);
}

void walk_dir() {
	outBuf = new emper::lib::ShardedFileBuffer(*Runtime::getRuntime(), STDOUT_FILENO);

	// Only search in regular files
	auto filter = [](const fs::directory_entry& dirent) { return dirent.is_regular_file(); };
	// Search the file
	auto fn = [](const fs::directory_entry& dirent) { search(dirent.path()); };

	switch (stealingMode) {
		case emper::StealingMode::child:
			emper::io::recursive_directory_walk(".", filter, fn, max_searching, max_walking);
			break;
		case emper::StealingMode::continuation:
			emper::io::fibril_recursive_directory_walk(".", filter, fn);
			break;
	}

	delete outBuf;
	Runtime::getRuntime()->initiateTermination();
}

static auto fssearch(const po::variables_map& vm) -> int {
	if (!vm.count("needle")) {
		std::cerr << "No 'needle' specified" << std::endl;
		return EXIT_FAILURE;
	}

	auto needleStr = vm["needle"].as<std::string>();
	needle = needleStr.c_str();
	needle_len = needleStr.length();

	if (vm.count("max-fibers")) {
		unsigned int maxFibersCount = vm["max-fibers"].as<unsigned int>();
		max_fibers = max_searching = max_walking = new emper::Semaphore(maxFibersCount);
	} else {
		if (vm.count("max-searchers")) {
			unsigned int maxSearchersCount = vm["max-searchers"].as<unsigned int>();
			max_searching = new emper::Semaphore(maxSearchersCount);
		}

		if (vm.count("max-walkers")) {
			unsigned int maxWalkersCount = vm["max-walkers"].as<unsigned int>();
			max_walking = new emper::Semaphore(maxWalkersCount);
		}
	}

	Runtime runtime;

	std::cerr << "Starting fsearch with stealingMode=" << stealingMode;
	if (max_fibers)
		std::cerr << " and " << max_fibers->getCount() << " fibers";
	else {
		if (max_searching) std::cerr << ", " << max_searching->getCount() << " searching fibers";
		if (max_walking) std::cerr << " and " << max_walking->getCount() << " walking fibers";
	}
	std::cerr << std::endl;

	auto* dirWalker = Fiber::from(walk_dir);
	runtime.scheduleFromAnywhere(*dirWalker);

	runtime.waitUntilFinished();

	if (max_searching == max_walking)
		delete max_fibers;
	else {
		delete max_searching;
		delete max_walking;
	}

	return EXIT_SUCCESS;
}

auto main(int argc, char* argv[]) -> int {
	po::options_description desc("Allowed options");
	// clang-format off
	desc.add_options()
		("help", "Show help")
		("needle", po::value<std::string>(), "The String to search for")
		("max-fibers", po::value<unsigned int>(), "Maximum number of fibers")
		("max-searchers", po::value<unsigned int>(), "Maximum number of file searching fibers")
		("max-walkers", po::value<unsigned int>(), "Maximum number of directory walking fibers")
		("stealing-mode", po::value<enum emper::StealingMode>(&stealingMode)->default_value(emper::StealingMode::child), "Stealing mode to use, either 'child' or 'continuation'")
		;
	// clang-format on

	// Make 'needle' a positional option.
	po::positional_options_description pos_desc;
	pos_desc.add("needle", -1);

	// clang-format off
	auto parse_result = po::command_line_parser(argc, argv)
		.options(desc)
		.positional(pos_desc)
		.run()
		;
	// clang-format on

	po::variables_map vm;
	po::store(parse_result, vm);
	po::notify(vm);

	if (vm.count("help")) {
		std::cout << desc << "\n";
		return EXIT_SUCCESS;
	}

	try {
		return fssearch(vm);
	} catch (const std::exception& e) {
		std::cerr << e.what();
		return 1;
	}
}
