// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#include "BinaryPrivateSemaphore.hpp"

#include <cassert>	// for assert

#include "Common.hpp"		// for unlikely
#include "Context.hpp"	// for Context
#include "Debug.hpp"		// for LOGDD
#include "Emper.hpp"		// for emper::DEBUG

void BinaryPrivateSemaphore::wait() {
	State state = bpsState.load();

	// Fast-path. We've already been signaled, no need to wait. Just
	// return.
	if (unlikely(state == signaled)) return;

	// If the BPS hasen't been signaled, then we have to be in initial
	// state.
	assert(state == initial);

	// If the binary signal was not yet signaled, then we need to
	// block here. That is, we perform a full context switch.
	blockedContext = Context::getCurrentContext();
	assert(blockedContext > (Context*)4096);
	block([this] {
		// N.B. wait() will have set the blockedContext at this point.
		State newState = blocked;
		State previousState = bpsState.exchange(newState);
		if (previousState == signaled) {
			if constexpr (emper::DEBUG) {
				// Reset the real signal state only in debug
				// builds. As it is not required to set the state here
				// for the correctnes of the synchronization primitive.
				newState = signaled;
				bpsState.store(newState, std::memory_order_relaxed);
			}
			unblock(blockedContext);
		}
	});
}

auto BinaryPrivateSemaphore::signalInternal() -> Context* {
	State currentState = bpsState.load();

	assert(currentState != signaled);

	if (currentState != initial) {
		// Fast path: If the state is not initial then someone is
		// waiting, because the state can't be signaled.
		if constexpr (emper::DEBUG) {
			State newState = signaled;
			// Reset the real signal state only in debug
			// builds. As it is not required to set the state here
			// for the correctnes of the synchronization primitive.
			bpsState.store(newState);
		}
		LOGDD("unblock in fast path, as we found a blocked context " << blockedContext);
		return blockedContext;
	}

	State newState = signaled;
	State previousState = bpsState.exchange(newState);

	if (unlikely(previousState == initial)) {
		// The state is still 'initial' which means the other
		// fiber hasn't called wait() yet.
		LOGDD("no unblock in slow path");
		return nullptr;
	}

	LOGDD("unblock in slow path");
	return blockedContext;
}
