// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#include "CountingPrivateSemaphore.hpp"

#include <cassert>	// for assert

#include "Context.hpp"	// for Context
#include "Debug.hpp"

CountingPrivateSemaphore::CountingPrivateSemaphore() : CountingPrivateSemaphore(0) {}

CountingPrivateSemaphore::CountingPrivateSemaphore(unsigned int counter)
		: counter(counter), blockedContext(nullptr) {}

void CountingPrivateSemaphore::incrementCounterByOne() { counter++; }

void CountingPrivateSemaphore::incrementCounter(unsigned int count) { counter += count; }

void CountingPrivateSemaphore::wait() {
	if (counter > 0) {
		Context* blockedContext = Context::getCurrentContext();
		block([this, blockedContext] {
			this->blockedContext = blockedContext;
			assert(blockedContext > (Context*)4096);
			if (this->getCounter() == 0) {
				Context* readyContext = this->blockedContext.exchange(nullptr);
				if (readyContext != nullptr) {
					unblock(readyContext);
				}
			}
		});
	}
}

auto CountingPrivateSemaphore::signalInternal() -> Context* {
	unsigned int oldCounter = counter.fetch_sub(1);
	assert(oldCounter >= 1);

	// If the counter is still non-zero after the decrement, somebody
	// else is responsible for scheduling the fiber.
	if (oldCounter > 1) {
		return nullptr;
	}

	if (blockedContext.load() != nullptr && counter == 0) {
		// Try to swap out a blocked context, it is fine if this
		// returns nullptr. In this case the block() function will
		// have won the race.
		Context* context = blockedContext.exchange(nullptr);
		EMPER_ASSERT_MSG(!context || context > (Context*)4096, "Unexpected context value: " << context);
		return context;
	}

	return nullptr;
}
