// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#include "Dispatcher.hpp"

#include <sched.h>

#include <atomic>
#include <chrono>

#include "Emper.hpp"
#include "Runtime.hpp"	// for Runtime
#include "WakeupStrategy.hpp"
#include "sleep_strategy/SemaphoreWorkerSleepStrategy.hpp"
#include "stats/Worker.hpp"

// Note that we initialize this with true to avoid counting workers
// that initially go to sleep as being woken up in vain.
thread_local bool Dispatcher::dispatchedWork = true;

auto Dispatcher::getDispatchLoop() -> func_t {
	return [this] { dispatchLoop(); };
}

void Dispatcher::dispatchLoopDoSleep() {
	runtime.maybeTerminateWorker();

	if constexpr (emper::WORKER_SLEEP) {
		dispatchLoopSleep();
	} else {
		sched_yield();
	}
}

void Dispatcher::dispatchLoopSleep() {
	auto& wakeupStrategy = runtime.wakeupStrategy;
	auto& workerSleepStrategy = runtime.workerSleepStrategy;
	auto& terminateWorkers = runtime.terminateWorkers;

	using clock = std::chrono::system_clock;
	std::chrono::time_point<clock> sleepStart;
	bool workerSlept = false;

	bool canWake;
	do {
		// Notify the wakeup strategy about our sleep attempt
		// TODO: We could take stats how often the wakeupStrategy decided
		// that the worker should not sleep, and how often this was the
		// right decission (based on dispatchedWork and a new thread_local sleepDenied).
		if (!wakeupStrategy.canSleep()) break;

		if constexpr (emper::STATS_WORKER_SLEEP) {
			if (!workerSlept) {
				if (!dispatchedWork) emper::stats::worker->wakeupInVainCount++;
				workerSlept = true;
				sleepStart = clock::now();
			}
		}

		workerSleepStrategy.sleep();

		// We always wakeup if the runtime is terminating
		canWake = wakeupStrategy.canWakeup() || terminateWorkers.load(std::memory_order_relaxed);
	} while (!canWake);

	if constexpr (emper::STATS_WORKER_SLEEP) {
		if (workerSlept) {
			auto sleepStop = clock::now();
			emper::stats::worker->recordSleep(sleepStart, sleepStop);
			// TODO: Not sure if we should set dispatchedWork to false regardless of the value of
			// firstSleep
			dispatchedWork = false;
		}
	}
}
