// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#include "Emper.hpp"

#include <cassert>
#include <cerrno>
#include <cstdint>
#include <stdexcept>
#include <utility>

#include "Common.hpp"
#include "Context.hpp"
#include "Fiber.hpp"
#include "Runtime.hpp"
#include "emper-common.h"
#include "emper-version.h"
#include "emper.hpp"
#include "io/Future.hpp"
#include "lib/LinuxVersion.hpp"
#include "lib/util.hpp"

namespace eu = emper::lib::util;

void async(Fiber* fiber) {
	assert(fiber != nullptr);
	Runtime* runtime = Runtime::getRuntime();
	runtime->schedule(*fiber);
}

void async(Fiber::fiber_fun_t function, void* arg) {
	Fiber* fiber = Fiber::from(std::move(function), arg);
	async(fiber);
}

void async(const Fiber::fiber_fun0_t& function) {
	Fiber* fiber = Fiber::from(function);
	async(fiber);
}

void async(Fiber::fiber_fun_t function, void* arg, workeraffinity_t* affinity) {
	Fiber* fiber = Fiber::from(std::move(function), arg, affinity);
	async(fiber);
}

void async(const Fiber::fiber_fun0_t& function, workeraffinity_t* affinity) {
	Fiber* fiber = Fiber::from(function, affinity);
	async(fiber);
}

namespace emper {

auto operator<<(std::ostream& out, const ContinuationStealingMode& cont_stealing_mode)
		-> std::ostream& {
	switch (cont_stealing_mode) {
		case ContinuationStealingMode::disabled:
			out << "disabled";
			break;
		case ContinuationStealingMode::locked:
			out << "locked";
			break;
		case ContinuationStealingMode::waitfree:
			out << "waitfree";
			break;
	}
	return out;
}

auto operator>>(std::istream& in, ContinuationStealingMode& cont_stealing_mode) -> std::istream& {
	std::string token;
	in >> token;
	if (token == "disabled") {
		cont_stealing_mode = ContinuationStealingMode::disabled;
	} else if (token == "locked") {
		cont_stealing_mode = ContinuationStealingMode::locked;
	} else if (token == "waitfree") {
		cont_stealing_mode = ContinuationStealingMode::waitfree;
	} else {
		in.setstate(std::ios_base::failbit);
	}

	return in;
}

auto operator<<(std::ostream& out,
								const ContinuationStealingMadviseStack& cont_stealing_madvise_stack)
		-> std::ostream& {
	switch (cont_stealing_madvise_stack) {
		case ContinuationStealingMadviseStack::no:
			out << "no";
			break;
		case ContinuationStealingMadviseStack::dontneed:
			out << "dontneed";
			break;
		case ContinuationStealingMadviseStack::free:
			out << "free";
			break;
	}
	return out;
}

auto operator>>(std::istream& in, ContinuationStealingMadviseStack& cont_stealing_madvise_stack)
		-> std::istream& {
	std::string token;
	in >> token;
	if (token == "no") {
		cont_stealing_madvise_stack = ContinuationStealingMadviseStack::no;
	} else if (token == "dontneed") {
		cont_stealing_madvise_stack = ContinuationStealingMadviseStack::dontneed;
	} else if (token == "free") {
		cont_stealing_madvise_stack = ContinuationStealingMadviseStack::free;
	} else {
		in.setstate(std::ios_base::failbit);
	}

	return in;
}

void printInfo(std::ostream& strm) {
	// clang-format off
	strm << std::boolalpha << "Efficient Massive-Parallelism Execution Realm (EMPER)" << std::endl
			 << "Version: " << getFullVersion() << std::endl
			 << "Debug: " << DEBUG << std::endl
			 << "IO: " << IO << std::endl
			 << "IO Stealing: " << IO_STEALING << std::endl
			 << "IO Lockless CQ: " << IO_LOCKLESS_CQ << std::endl
			 << "Set Affinity on Block: " << SET_AFFINITY_ON_BLOCK << std::endl
			 << "Work-Stealing Queue Default: " << TOSTRING(EMPER_WS_QUEUE_DEFAULT_TYPE) << std::endl
			 << "Work-Stealing Queue Scheduler: " << TOSTRING(EMPER_WS_QUEUE_SCHEDULER_TYPE) << std::endl
			 << "Continuation-Stealing: " << CONTINUATION_STEALING_MODE << std::endl
			 << "Continuation-Stealing Madvise Stack: " << CONTINUATION_STEALING_MADVISE_STACK << std::endl
			 << "Context Size: " << eu::bytesToHumanReadableString(Context::CONTEXT_SIZE) << std::endl
			 << "Context Manager with Memory Manager: " << CONTEXT_MANAGER_WITH_MEMORY_MANAGER << std::endl
			 << "Memory Manager Victim Percentage: " << MEMORY_MANAGER_VICTIM_PERCENTAGE << std::endl
			 << "Stack Guard Page: " << STACK_GUARD_PAGE << std::endl
			 << "Assume Cache Line Size: " << ASSUME_CACHE_LINE_SIZE << std::endl
			 << "Stats: " << STATS << std::endl
			 << "Worker Sleep: " << WORKER_SLEEP << std::endl
			 << "Overflow Queue: " << OVERFLOW_QUEUE << std::endl
		;
	// clang-format on
}

const bool IO_MUST_INVALIDATE_BROKEN_CHAIN = EMPER_LINUX_LT("5.15.0");
const bool IO_FORGOTTEN_FUTURES_SKIP_CQE = EMPER_LINUX_GE("5.17");

auto getFullVersion() -> std::string { return EMPER_FULL_VERSION; }

static void ensure_no_current_runtime() {
	Runtime* runtime = Runtime::getRuntime();
	if (runtime) {
		throw std::runtime_error("Runtime already initialized");
	}
}

void init_runtime() {
	ensure_no_current_runtime();

	// NOLINTNEXTLINE(clang-analyzer-cplusplus.NewDeleteLeaks)
	new Runtime();
	// NOLINTNEXTLINE(clang-analyzer-cplusplus.NewDeleteLeaks)
}

void init_runtime(workerid_t worker_count) {
	ensure_no_current_runtime();

	// NOLINTNEXTLINE(clang-analyzer-cplusplus.NewDeleteLeaks)
	new Runtime(worker_count);
	// NOLINTNEXTLINE(clang-analyzer-cplusplus.NewDeleteLeaks)
}

void destroy_runtime() {
	Runtime* runtime = Runtime::getRuntime();
	if (!runtime) {
		throw std::runtime_error("No runtime to destroy");
	}

	delete runtime;
}

void yield() {
	Runtime* runtime = Runtime::getRuntime();
	runtime->yield();
}

auto sleep(unsigned int seconds) -> bool {
	if constexpr (!emper::IO) {
		DIE_MSG("sleep requires emper::io");
	}

	emper::io::AlarmFuture::Timespec ts = {.tv_sec = seconds, .tv_nsec = 0};
	emper::io::AlarmFuture alarm(ts);
	int32_t res = alarm.submitAndWait();

	return res == -ETIME;
}

}	 // namespace emper
