// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#pragma once

#include <pthread.h>	// for pthread_t
#include <signal.h>

#include <atomic>
#include <chrono>
#include <cstdint>		 // for intptr_t
#include <functional>	 // for function
#include <iostream>
#include <memory>
#include <mutex>
#include <optional>
#include <random>
#include <string>
#include <vector>	 // for vector

#include "CallerEnvironment.hpp"
#include "Common.hpp"
#include "Debug.hpp"
#include "FiberHint.hpp"
#include "FiberSource.hpp"
#include "Scheduler.hpp"	// for Scheduler
#include "WakeupStrategy.hpp"
#include "Worker.hpp"
#include "emper-common.h"			 // for workerid_t
#include "lib/sync/Latch.hpp"	 // for Latch
#include "lib/sync/Semaphore.hpp"
#include "sleep_strategy/WorkerSleepStrategy.hpp"

enum class LogSubsystem;
class RuntimeBuilder;
class ContextManager;
class Dispatcher;
class Fiber;
struct NextFiberResult;
class RuntimeStrategy;
class RuntimeStrategyStats;
template <typename D>
class WorkerLocalData;

namespace emper {
namespace io {
class GlobalIoContext;
class IoContext;
}	 // namespace io
namespace lib {
class ShardedFileBuffer;
}
namespace log {
class LogBuffer;
}
namespace sleep_strategy {
class PipeSleepStrategy;
}
namespace stats {
class Worker;
class FromAnywhere;
}	 // namespace stats
}	 // namespace emper

using emper::io::GlobalIoContext;
using emper::io::IoContext;
using emper::sleep_strategy::WorkerSleepStrategy;

class Runtime : public Logger<LogSubsystem::RUNTI> {
 public:
	using NewWorkerHook = std::function<void(workerid_t)>;

	using StrategyFactory = std::function<RuntimeStrategy*(Runtime&)>;

	const std::chrono::time_point<std::chrono::system_clock> constructionTime =
			std::chrono::system_clock::now();

 private:
	static std::mutex currentRuntimeMutex;
	static Runtime* currentRuntime;

	const workerid_t workerCount;

	std::vector<NewWorkerHook> newWorkerHooks;

	Latch workerLatch;
	Latch firstWorkerThreadExitLatch;
	Latch secondWorkerThreadExitLatch;

	RuntimeStrategy* const strategy;

 public:
	Scheduler& scheduler;

 private:
	Dispatcher& dispatcher;
	ContextManager& contextManager;
	pthread_t* threads;
	Worker** workers;
	GlobalIoContext* globalIo = nullptr;
	std::vector<IoContext*> ioContexts;
	emper::lib::sync::Semaphore ioReadySem;

	std::default_random_engine randomEngine;
	std::uniform_int_distribution<unsigned int> uniformIntDistribution;

	std::atomic<bool> terminateWorkers = false;

	bool threadsRunning = false;

	std::shared_ptr<RuntimeStrategyStats> strategyLastStats;

	WorkerLocalData<emper::stats::Worker>* workerStats;
	emper::stats::FromAnywhere* fromAnywhereStats;

	emper::log::LogBuffer* logBuffer = nullptr;

	workerid_t pinningOffset = 0;

	auto workerLoop(Worker* worker) -> void*;

	WorkerSleepStrategy workerSleepStrategy;
	emper::WakeupStrategy wakeupStrategy;

	// Per runtime-system stats
	ALIGN_TO_CACHE_LINE std::atomic<size_t> blockedContextCount{0};

	static StrategyFactory DEFAULT_STRATEGY_FACTORY;

	static void printLastRuntimeStats(std::ostream& out = std::cout);

	static auto shouldPinWorkers() -> bool;
	static auto getDefaultPinningOffset() -> workerid_t;

	static bool guardPageTouchCausesExitSuccess;

	static struct sigaction oldact;

	static void sigSegvHandler(int sig, siginfo_t* info, void* ucontext);

 protected:
	void addNewWorkerHook(const std::function<void(workerid_t)>& hook) {
		newWorkerHooks.push_back(hook);
	};

	[[nodiscard]] inline auto getWorkerSleepStrategy() -> WorkerSleepStrategy& {
		return workerSleepStrategy;
	}

	template <CallerEnvironment callerEnvironment>
	inline void wakeupSleepingWorkers(emper::FiberHint hint) {
		LOGD("Wake sleepers from " << callerEnvironment << " cause new work: " << hint);
		switch (hint.getSource()) {
			// We scheduled to an mpsc queue where the fiber lives exclusively -> wake up the
			// specific worker if necessary
			case emper::FiberSource::mpscQueue: {
				workerSleepStrategy.notifySpecific<callerEnvironment>(hint.getWorker());
			} break;

			// We scheduled to a work-stealing or the anywhereQueue, where anyone can
			// access this fiber -> wake up anyone if appropriate
			case emper::FiberSource::local:
			case emper::FiberSource::anywhereQueue: {
				workerid_t wakeupCount = wakeupStrategy.getWakeupCount<callerEnvironment>();
				if (wakeupCount) {
					workerSleepStrategy.notifyMany<callerEnvironment>(wakeupCount);
				}
			} break;

			default:
				DIE_MSG("Unexpected FiberSource encountered during wakeupSleepingWorkers");
		}
	}

	void maybeTerminateWorker();

	auto nextFiber() -> std::optional<NextFiberResult>;

 public:
	Runtime() : Runtime(getDefaultWorkerCount()) {}

	Runtime(workerid_t workerCount) : Runtime(workerCount, DEFAULT_STRATEGY_FACTORY) {}

	Runtime(const StrategyFactory& strategyFactory)
			: Runtime(getDefaultWorkerCount(), strategyFactory) {}

	Runtime(workerid_t workerCount, const StrategyFactory& strategyFactory,
					unsigned int seed = std::random_device()())
			: Runtime(workerCount, std::vector<NewWorkerHook>(), shouldPinWorkers(),
								getDefaultPinningOffset(), strategyFactory, seed) {}

	Runtime(workerid_t workerCount, const std::vector<NewWorkerHook>& newWorkerHooks, bool pinWorkers,
					workerid_t pinningOffset, const StrategyFactory& strategyFactory, unsigned int seed);

	~Runtime();

	inline void schedule(Fiber& fiber) { scheduler.schedule(fiber); }

	inline void schedule(Fiber** fibers, unsigned count) { scheduler.schedule(fibers, count); }

	inline void scheduleFromAnywhere(Fiber& fiber) { scheduler.scheduleFromAnywhere(fiber); }

	inline void scheduleFromAnywhere(Fiber** fibers, unsigned count) {
		scheduler.scheduleFromAnywhere(fibers, count);
	}

	/**
	 * @brief Schedule a fiber on a specific worker
	 *
	 * For now scheduleOn is only available from within th EMPER runtime.
	 */
	inline void scheduleOn(Fiber& fiber, workerid_t workerId) {
		scheduler.scheduleOn(fiber, workerId);
	}
	void yield();

	// https://stackoverflow.com/a/3747462/194894
	static inline auto rand() -> int { return Worker::rand(); }

	static inline auto getWorkerId() -> workerid_t { return Worker::getCurrentWorkerId(); }

	[[nodiscard]] inline auto cpuToWorkerId(unsigned cpu) const -> workerid_t {
		return cpu - pinningOffset;
	}

	[[nodiscard]] inline auto getWorkerCount() const -> workerid_t { return workerCount; }

	static inline auto getRuntime() -> Runtime* { return currentRuntime; }

	static void setGuardPageTouchCausesExitSuccess() { guardPageTouchCausesExitSuccess = true; }

	inline auto getContextManager() -> ContextManager& { return contextManager; }

	inline auto getStrategy() -> RuntimeStrategy& { return *strategy; }

	[[nodiscard]] inline auto isTerminating() const -> bool { return terminateWorkers; }

	void initiateTermination();

	void waitUntilFinished();

	void initiateAndWaitUntilTermination();

	void printStats(std::ostream& out = std::cout, bool detailed = false);

	void executeAndWait(std::function<void()> f);

	static auto getDefaultWorkerCount() -> workerid_t;

	friend class AbstractWorkStealingScheduler;
	friend class WsDispatcher;
	friend class LawsDispatcher;
	template <LogSubsystem>
	friend class Blockable;
	friend RuntimeBuilder;
	friend ContextManager;
	friend Scheduler;
	friend Dispatcher;
	friend GlobalIoContext;
	friend IoContext;
	template <typename T, intptr_t WS_QUEUE_SIZE, size_t WORKER_EXCLUSIVE_QUEUE_SIZE>
	friend class MemoryManager;
	template <typename>
	friend class WorkerLocalData;
	friend class Fibril;
	friend class emper::sleep_strategy::PipeSleepStrategy;
	friend class emper::lib::ShardedFileBuffer;

	friend void emper::log::log(const std::string& prefix, const std::string& message);
};
