// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#include "io/IoContext.hpp"

#include <liburing.h>						// for io_uring_submit, io_uring_get_sqe
#include <liburing/io_uring.h>	// for io_uring_cqe, io_uring_params, IORI...
#include <numa.h>
#include <sched.h>
#include <sys/eventfd.h>
#include <unistd.h>

#include <algorithm>
#include <array>
#include <atomic>		// for atomic, __atomic_base
#include <cassert>	// for assert
#include <cerrno>		// for errno, ECANCELED, EBUSY, EAGAIN, EINTR
#include <cstring>	// for memset
#include <optional>
#include <ostream>	// for basic_osteram::operator<<, operator<<
#include <string>
#include <vector>

#include "BinaryPrivateSemaphore.hpp"
#include "CallerEnvironment.hpp"	// for CallerEnvironment, EMPER, ANYWHERE
#include "Common.hpp"							// for unlikely, DIE_MSG_ERRNO, DIE_MSG
#include "Debug.hpp"							// for LOGD
#include "Emper.hpp"							// for DEBUG, IO_URING_SQPOLL
#include "Runtime.hpp"
#include "emper-common.h"
#include "emper-config.h"
#include "io/Future.hpp"
#include "io/GlobalIoContext.hpp"
#include "io/Stats.hpp"	 // for Stats, nanoseconds
#include "io/SubmitActor.hpp"
#include "lib/LinuxVersion.hpp"
#include "lib/TaggedPtr.hpp"
#include "lib/env.hpp"

using emper::lib::TaggedPtr;

namespace emper::io {

thread_local IoContext *IoContext::workerIo = nullptr;

auto IoContext::prepareFutureChain(Future &future, unsigned chain_length) -> unsigned {
	unsigned total_chain_length = chain_length;

	if (unlikely(chain_length > *this->ring.sq.kring_entries)) {
		DIE_MSG("Future chain is to long for the io_uring's SQ");
	}

	if (future.dependency) {
		total_chain_length = prepareFutureChain(*future.dependency, chain_length + 1);
	}

	if constexpr (emper::DEBUG) {
		reqs_in_uring++;
		uringFutureSet.insert(&future);
	}

	struct io_uring_sqe *sqe = getSqe();

	future.prepareSqe(*this, sqe);

	// remember the prepared sqe to invalidate it if it was not properly submitted
	preparedSqes.push_back(sqe);

	void *user_data = createFutureTag(future);

	// We should start a new Fiber executing callback on completion
	if (future.callback) {
		LOGD("prepare " << future << " Callback " << future.callback);
		// set the Callback's affinity to the submitting worker
		if (worker) future.callback->affinity = worker->getWorkerId();
	}

	// Appropriately tag the request so that it can be canceled and we can
	// handle its completion.
	io_uring_sqe_set_data(sqe, user_data);

	if (future.isForgotten()) {
		// Do not generate cqes for forgotten futures this reduces the kernel
		// and user space overhead for cqes we do not need in the first place.
		// IOSQE_CQE_SKIP_SUCCESS is supported since Linux 5.17
		if (IO_FORGOTTEN_FUTURES_SKIP_CQE) sqe->flags |= IOSQE_CQE_SKIP_SUCCESS;
	}

	// If we use a single io_uring and a SubmitActor the submission and preparation
	// of futures are decoupled.
	// The Future may be used in a Fiber concurrently to the SubmitActor submitting it
	// to the single io_uring.
	// We must ensure that the Future is not dropped before it is prepared.
	//
	// Future::submitAndForget does this by waiting on the BPS of the Future
	// which will not be signaled on completion because the Future is forgotten.
	// We must signal the preparation of the forgotten future and allow its possible
	// destruction.
	//
	// And the preparation of Futures with callbacks is now awaited during cancel
	// which runs in ~Future which ensures that they can be dropped safely.
	if (future.isForgotten() || future.callback) {
		future.sem.signal();
	}

	if (future.isDependency()) {
		LOGD("Prepare " << future << " as a dependency");
		sqe->flags |= IOSQE_IO_LINK;
	}

	return total_chain_length;
}

template <CallerEnvironment callerEnvironment>
auto IoContext::submitPreparedSqesAndWait(unsigned wait_nr) -> unsigned {
	// submit the Future to the io_uring
	int submitted = io_uring_submit_and_wait(&ring, wait_nr);

	// Actually I don't know how "unlikely" this is
	if (unlikely(submitted == -EBUSY)) {
		// We can't submit our sqe because the CQ is full
		// in worker thread -> reapCompletions
		// in globalCompleter thread -> TODO: deterministically handle global full CQ
		// for now hope a jam does not happen or will solve itself
		LOGD("io_submit returned EBUSY trying to submit " << io_uring_sq_ready(&ring)
																											<< " in addition to " << reqs_in_uring);

		if constexpr (callerEnvironment == CallerEnvironment::ANYWHERE) {
			// we are not busy looping in the globalIo
			DIE_MSG("Future dropped because global completer SQ is full");
		}

		TIME_NS(
				{
					do {
						reapAndScheduleCompletions<callerEnvironment>();
					} while ((submitted = io_uring_submit(&ring)) == -EBUSY);
				},
				stats.record_io_submit_full_cq);
	}

	if (unlikely(submitted < 0)) {
		errno = -submitted;
		DIE_MSG_ERRNO("io_uring_submit failed");
	}

	return static_cast<unsigned>(submitted);
}

template <CallerEnvironment callerEnvironment>
void IoContext::cancelUnsubmittedChainParts(Future &chain) {
	unsigned unsubmitted = io_uring_sq_ready(&ring);
	Future *unsubmittedFuture = &chain;

	for (unsigned i = 0; i < unsubmitted; ++i) {
		// we prepare sqes for future from the chain end to the head
		// A -> B -> C will result in CQ [C, B, A] and preparedSqes [C, B, A]
		// invalidate/cancel N sqes/Futures from the back
		struct io_uring_sqe *sqe = preparedSqes.back();
		preparedSqes.pop_back();
		// invalidate the prepared sqe
		io_uring_prep_nop(sqe);
		io_uring_sqe_set_data(sqe, nullptr);

		// cancel the according future
		unsubmittedFuture->recordCompletion(stats, -ECANCELED);
		if constexpr (callerEnvironment == EMPER) {
			unsubmittedFuture->complete(-ECANCELED);
		} else {
			unsubmittedFuture->completeFromAnywhere(-ECANCELED);
		}
		unsubmittedFuture = chain.dependency;
	}

	// submit all now invalidated sqes
#ifdef NDEBUG
	ATTR_UNUSED
#endif
	unsigned invalidated = submitPreparedSqes<callerEnvironment>();
	assert(invalidated == unsubmitted);
}

// show the compiler our template incarnations
template void IoContext::cancelUnsubmittedChainParts<CallerEnvironment::EMPER>(Future &future);
template void IoContext::cancelUnsubmittedChainParts<CallerEnvironment::ANYWHERE>(Future &future);

template <CallerEnvironment callerEnvironment>
void IoContext::submitAndWait(Future &future, unsigned wait_nr) {
	LOGD("submitting " << future << (future.dependency ? " and it's dependencies" : ""));
	unsigned prepared = prepareFutureChain(future);

	// submit the Future to the io_uring
	int submitted = submitPreparedSqesAndWait<callerEnvironment>(wait_nr);

	// We submitted some Futures to the io_uring
	// Because we submit every Future right away multiple prepared sqes can only
	// occur for Future chains.
	// If we could not submit the whole chain cancel all non submitted Futures
	// because we can not guaranty the soundness of the chain have to invalidate
	// non submitted but prepared sqes
	// req1 -> invalid_req -> req3
	// will submit only 2 instead of all 3 prepared sqes leaving a sqe for req3
	// in the SQ but unsubmitted
	// See: https://github.com/axboe/liburing/issues/186
	// https://github.com/axboe/liburing/issues/92
	//
	// Since kernel commit cf10960426515 io_uring does no longer submit parts of a
	// broken chains. This means that all sqes in a chain except the broken one will
	// result in cqes with result -ECANCELD and the invalid one will
	// generate a cqe with the appropriate error code

	// When 5.15 is released with
	// https://lore.kernel.org/io-uring/180ec124-79b1-2274-4570-9b0d6620d512@linux.alibaba.com/T/#t
	// the kernel will consume the whole broken chain and we
	// don't need to manually invalidate not submitted sqes.
	// This allows SQPOLL to work for invalid chains.
	if (unlikely(static_cast<unsigned>(submitted) < prepared) &&
			emper::IO_MUST_INVALIDATE_BROKEN_CHAIN) {
		cancelUnsubmittedChainParts<callerEnvironment>(future);
	}

	// we have submitted all our prepared sqes
	preparedSqes.clear();
}

// show the compiler our template incarnations
template void IoContext::submitAndWait<CallerEnvironment::EMPER>(Future &future, unsigned wait_nr);
template void IoContext::submitAndWait<CallerEnvironment::ANYWHERE>(Future &future,
																																		unsigned wait_nr);

/**
 * Configurable memory order for the atomic operations
 */
static constexpr auto LL_READ_MEM_ORDER =
		emper::IO_LOCKLESS_MEMORY_ORDER == emper::IoLocklessMemoryOrder::weak
				? std::memory_order_acquire
				: std::memory_order_seq_cst;
static constexpr auto LL_WRITE_MEM_ORDER =
		emper::IO_LOCKLESS_MEMORY_ORDER == emper::IoLocklessMemoryOrder::weak
				? std::memory_order_release
				: std::memory_order_seq_cst;

template <CallerEnvironment callerEnvironment>
auto IoContext::tryReapCompletionWaitFree(Fiber **continuation) -> emper::StealingResult {
	Completion completion;

	struct io_uring_cq *cq = &ring.cq;
	const unsigned mask = *cq->kring_mask;
	auto *atail = reinterpret_cast<std::atomic<unsigned> *>(cq->ktail);
	auto *ahead = reinterpret_cast<std::atomic<unsigned> *>(cq->khead);

	// NOTE: just using head for the CAS introduces a possible ABA problem
	//       if the unsigned head counter overflows during the read and the CAS.

	// Load possibly concurrently used userspace written head pointer
	unsigned head = ahead->load(LL_READ_MEM_ORDER);
	// Load concurrently used kernel written tail pointer
	unsigned tail = atail->load(LL_READ_MEM_ORDER);

	// NOTE: This number may already be wrong during its calculation
	unsigned ready = tail - head;

	if (!ready) return emper::StealingResult::Empty;

	const struct io_uring_cqe *cqe = &cq->cqes[head & mask];
	void *cqe_data = io_uring_cqe_get_data(cqe);

	// Only the owner is allowed to reap new work notifications
	if constexpr (callerEnvironment != CallerEnvironment::OWNER) {
		const TaggedPtr tptr(cqe_data);
		const auto tag = static_cast<PointerTags>(tptr.getTag());
		if (tag == PointerTags::NewWorkNotification) {
			// don't consume the new work notification
			return emper::StealingResult::Empty;
		}
	}

	completion.first = cqe->res;
	completion.second = cqe_data;

	if (!ahead->compare_exchange_strong(head, head + 1, LL_WRITE_MEM_ORDER, LL_READ_MEM_ORDER))
		return emper::StealingResult::LostRace;

	trackReqsInUring(-1);

	// Reaps done by other worker threads are counted in the AbstractWorkStealingStats
	// as stolenIoFibers.
	stats.record_reaps<callerEnvironment>(1);

	getContinuationsFromCompletions<callerEnvironment>(&completion, 1, continuation);
	return emper::StealingResult::Stolen;
}

// show the compiler our template incarnations
template auto IoContext::tryReapCompletionWaitFree<CallerEnvironment::OWNER>(Fiber **continuation)
		-> emper::StealingResult;
template auto IoContext::tryReapCompletionWaitFree<CallerEnvironment::EMPER>(Fiber **continuation)
		-> emper::StealingResult;
template auto IoContext::tryReapCompletionWaitFree<CallerEnvironment::ANYWHERE>(
		Fiber **continuation) -> emper::StealingResult;

template <CallerEnvironment callerEnvironment, unsigned toReap>
auto IoContext::reapCompletionsLockless(Fiber **continuations) -> unsigned {
	Completion reapedCompletions[toReap];	 // NOLINT(modernize-avoid-c-arrays)

	struct io_uring_cq *cq = &ring.cq;
	const unsigned mask = *cq->kring_mask;
	auto *atail = reinterpret_cast<std::atomic<unsigned> *>(cq->ktail);
	auto *ahead = reinterpret_cast<std::atomic<unsigned> *>(cq->khead);

	// NOTE: just using head for the CAS introduces a possible ABA problem
	//       if the unsigned head counter overflows during the read and the CAS.

	// Load possibly concurrently used userspace written head pointer
	unsigned head = ahead->load(LL_READ_MEM_ORDER);
	unsigned count;
	do {
		// Load concurrently used kernel written tail pointer
		unsigned tail = atail->load(LL_READ_MEM_ORDER);

		// NOTE: This number may already be wrong during its calculation
		unsigned ready = tail - head;

		if (!ready) return 0;

		count = std::min(toReap, ready);

		for (unsigned i = 0; i < count; ++i) {
			const struct io_uring_cqe *cqe = &cq->cqes[(head + i) & mask];
			void *cqe_data = io_uring_cqe_get_data(cqe);

			// Only the owner is allowed to reap new work notifications
			if constexpr (callerEnvironment != CallerEnvironment::OWNER) {
				const TaggedPtr tptr(cqe_data);
				const auto tag = static_cast<PointerTags>(tptr.getTag());
				if (tag == PointerTags::NewWorkNotification) {
					// don't consume the new work notification
					if (i == 0) return 0;
					// Since i starts at 0 using i as count is correct.
					// If i = 1 this means we are at the second cqe and
					// count = i = 1 will consume only the first cqe.
					count = i;
					break;
				}
			}

			auto &reapedCompletion = reapedCompletions[i];
			reapedCompletion.first = cqe->res;
			reapedCompletion.second = cqe_data;
		}

		// TODO: think about the correct memory ordering constraints
	} while (
			!ahead->compare_exchange_weak(head, head + count, LL_WRITE_MEM_ORDER, LL_READ_MEM_ORDER));

	LOGD("got " << count << " cqes from worker " << worker->getWorkerId() << "'s io_uring");

	trackReqsInUring(-static_cast<long>(count));

	// Reaps done by other worker threads are counted in the AbstractWorkStealingStats
	// as stolenIoFibers.
	stats.record_reaps<callerEnvironment>(count);

	return getContinuationsFromCompletions<callerEnvironment>(reapedCompletions, count,
																														continuations);
}

// Show the compiler our template incarnations
// Owner reaps all available cqes
template auto IoContext::reapCompletionsLockless<CallerEnvironment::OWNER,
																								 IoContext::CQE_BATCH_COUNT>(Fiber **continuations)
		-> unsigned;

// Other worker steals a single cqe
template auto IoContext::reapCompletionsLockless<CallerEnvironment::EMPER, 1>(Fiber **continuations)
		-> unsigned;

// Completer reaps a single cqe
template auto IoContext::reapCompletionsLockless<CallerEnvironment::ANYWHERE, 1>(
		Fiber **contiunations) -> unsigned;
// Completer reaps a single cqe
template auto IoContext::reapCompletionsLockless<CallerEnvironment::ANYWHERE,
																								 IoContext::CQE_BATCH_COUNT>(Fiber **contiunations)
		-> unsigned;

#define UNLOCK_IF_NEEDED()                       \
	do {                                           \
		if constexpr (needsCqLock) cq_lock.unlock(); \
	} while (false);

template <CallerEnvironment callerEnvironment, unsigned toReap>
auto IoContext::reapCompletionsLocked(Fiber **continuations) -> unsigned {
	// TODO: should only the owner possibly rereap?
	constexpr bool checkForRereap = callerEnvironment == CallerEnvironment::OWNER;
	unsigned reReapCount = 0;

	// array to store cqe* needed by liburing
	struct io_uring_cqe *cqes[toReap];	// NOLINT(modernize-avoid-c-arrays)

	// array to store seen cqes to make the critical section
	// where cq_lock is held as small as possible
	Completion reapedCompletions[toReap];	 // NOLINT(modernize-avoid-c-arrays)

// this label is not used for callerEnvironment::ANYWHERE and thus has to be
// annotated with ATTR_UNUSED
reap_cqes:
	ATTR_UNUSED;

	// Number of actual continuation fibers resulting from the reaped CQEs
	unsigned continuationsCount = 0;

	if constexpr (needsCqLock) {
		// The Owner always takes the lock to reap all completions and especially
		// new work notifications.
		if constexpr (callerEnvironment == CallerEnvironment::OWNER) {
			cq_lock.lock();
		} else {
			// Someone else is currently reaping completions
			if (unlikely(!cq_lock.try_lock())) {
				LOGD("unsuccessful try_lock from " << callerEnvironment);
				return 0;
			}
		}
	}

	unsigned count = io_uring_peek_batch_cqe(&ring, cqes, toReap);
	if (!count) {
		UNLOCK_IF_NEEDED()

		if constexpr (checkForRereap) {
			goto check_for_rereap;
		}

		return 0;
	}

	for (unsigned i = 0; i < count; ++i) {
		struct io_uring_cqe *cqe = cqes[i];
		void *cqe_data = io_uring_cqe_get_data(cqe);

		auto &reapedCompletion = reapedCompletions[i];
		reapedCompletion.first = cqe->res;
		reapedCompletion.second = cqe_data;

		// Do not reap newWorkNotifications if we are no the owner
		if constexpr (callerEnvironment != CallerEnvironment::OWNER) {
			const auto tag = static_cast<PointerTags>(TaggedPtr(cqe_data).getTag());
			if (tag == PointerTags::NewWorkNotification) {
				UNLOCK_IF_NEEDED()
				return 0;
			}
		}
	}

	io_uring_cq_advance(&ring, count);
	UNLOCK_IF_NEEDED()

	LOGD("got " << count << " cqes from worker " << worker->getWorkerId() << "'s io_uring");

	trackReqsInUring(-static_cast<long>(count));

	// Reaps done by other worker threads are counted in the AbstractWorkStealingStats
	// as stolenIoFibers.
	stats.record_reaps<callerEnvironment>(count);

	continuationsCount =
			getContinuationsFromCompletions<callerEnvironment>(reapedCompletions, count, continuations);

	// A naive try lock protecting a worker's IoContext's cq is racy.
	// While a worker is holding the lock additional completions could arrive
	// which the worker does not observe because it could be already finished iterating.
	// In the case that the worker still holds the lock preventing others
	// from reaping the additional completions we have a lost wakeup possibly leading
	// to a completely sleeping runtime with runnable completions in a worker's IoContext.

	// To prevent this race we check the CQ again for new cqes after we already
	// dropped the CQ lock and possibly reap again.

	// Check if we missed new cqes.
check_for_rereap:
	ATTR_UNUSED;
	if constexpr (checkForRereap) {
		bool reReap = cqeCount() != 0;
		if (reReap) {
			// schedule all already collected continuation fibers
			runtime.schedule(continuations, continuationsCount);

			reReapCount++;

			LOGD("Re-Reaping completions");
			goto reap_cqes;
		}

		stats.record_reReapCount(reReapCount);
	}

	return continuationsCount;
}

// Show the compiler our template incarnations
// OWNER reaps all available cqes
template auto IoContext::reapCompletionsLocked<CallerEnvironment::OWNER,
																							 IoContext::CQE_BATCH_COUNT>(Fiber **continuations)
		-> unsigned;

// Other worker steals a single cqe
template auto IoContext::reapCompletionsLocked<CallerEnvironment::EMPER, 1>(Fiber **continuations)
		-> unsigned;

// Completer reaps a single cqe
template auto IoContext::reapCompletionsLocked<CallerEnvironment::ANYWHERE, 1>(
		Fiber **contiunations) -> unsigned;

// Completer reaps all available cqes
template auto IoContext::reapCompletionsLocked<CallerEnvironment::ANYWHERE,
																							 IoContext::CQE_BATCH_COUNT>(Fiber **contiunations)
		-> unsigned;

IoContext::IoContext(Runtime &runtime, size_t uring_entries, unsigned unboundedIowMax)
		: runtime(runtime) {
	const bool isWorkerIo = runtime.globalIo != nullptr;
	struct io_uring_params params;
	memset(&params, 0, sizeof(params));

	// Setup sq polling
	if constexpr (emper::IO_SQ_POLLER != IoSqPoller::off) {
		// Using a poller needs at least linux 5.15 to safely handle
		// broken chains see: 43b8dc163f985341d2a2a24d1bfe0e6bb939f880.
		if (emper::IO_MUST_INVALIDATE_BROKEN_CHAIN) {
			DIE_MSG("Invalidating broken chains is not possible when using sqpoll");
		}
		params.flags |= IORING_SETUP_SQPOLL;

		// We share poller threads and may need to attach to someone
		if constexpr (emper::IO_SQ_POLLER != IoSqPoller::each) {
			int pollerRingFd = -1;

			// We use a single poller -> attach to the global io_uring
			if constexpr (emper::IO_SQ_POLLER == IoSqPoller::one) {
				if (isWorkerIo) {
					pollerRingFd = runtime.globalIo->ring.ring_fd;
				}
				// One sq poller thread per numa node
				// The CPU with the lowest id in a node is responsible to create a poller.
				// All other CPUs in the same node will attach to this io_uring
			} else if constexpr (emper::IO_SQ_POLLER == IoSqPoller::numa) {
				// get our numa node
				unsigned cpu, node;
				int err = getcpu(&cpu, &node);
				if (unlikely(err)) {
					DIE_MSG_ERRNO("getcpu failed");
				}

				// get all CPUs in our node
				struct bitmask *mask = numa_allocate_cpumask();
				if (unlikely(!mask)) {
					DIE_MSG_ERRNO("numa_allocate_cpumask failed");
				}

				err = numa_node_to_cpus(static_cast<int>(node), mask);
				if (unlikely(err)) {
					DIE_MSG_ERRNO("getcpu failed");
				}

				// find the lowest cpu in the current node
				for (unsigned i = 0; i < mask->size; ++i) {
					if (!numa_bitmask_isbitset(mask, i)) {
						continue;
					}

					LOGD("Found first CPU " << i << " in node " << node << " from CPU " << cpu);
					IoContext *pollerIo = runtime.ioContexts[runtime.cpuToWorkerId(i)];
					if (pollerIo) {
						pollerRingFd = pollerIo->ring.ring_fd;
					}
					break;
				}
			}

			// attach to the used poller's ring fd
			if (pollerRingFd != -1) {
				params.flags |= IORING_SETUP_ATTACH_WQ;
				params.wq_fd = pollerRingFd;
			}
		}
	}

	auto ret = io_uring_queue_init_params(uring_entries, &ring, &params);
	if (ret < 0) {
		DIE_MSG_ERRNO("io_uring_queue_init failed");
	}
	LOGD("Ring fd is " << ring.ring_fd);

	if (!(params.features & IORING_FEAT_NODROP)) {
		LOGW("kernel das not support IORING_FEAT_NODROP. We may loose IO requests");
	}

	if constexpr (emper::IO_SQ_POLLER != IoSqPoller::off) {
		if (!(params.features & IORING_FEAT_SQPOLL_NONFIXED)) {
			DIE_MSG("io_uring_sqpoll defined but kernel does not support IORING_FEAT_SQPOLL_NONFIXED");
		}
	}

	// Detect IORING_REGISTER_RING_FDS support
	// Liburing hides the details of calling io_uring_enter with the index of the
	// registered io_uring fd instead of the fd itself.
	// We only have to register our io_uring's fd.
	// Added with Linux: e7a6c00dc77aedf27a601738ea509f1caea6d673

	// Do not register the ring fd on the global io_uring because it is created on
	// the main thread and afterwards used on the completer thread.
	// Registering on the main thread breaks the assumption done in
	// liburing that the registered offset is valid on the using thread.
	// Registering the global io_uring is also not possible when using SINGLE_URING.
	if (EMPER_LINUX_GE("5.18") && isWorkerIo) {
		ret = io_uring_register_ring_fd(&ring);
		if (unlikely(ret < 0)) {
			errno = -ret;
			DIE_MSG_ERRNO("registering our io_uring fd " << ring.ring_fd << " failed");
		}
	}

	// Limit the number of unbounded iow thread created by this io_uring
	// By default the number of unbounded iow worker threads is RLIMIT_NPROC
	// which may be excessive.
	// IORING_REGISTER_IOWQ_MAX_WORKERS sets a maximum value for bounded and unbounded
	// worker creation per NUMA node and thread.
	// See:
	// https://blog.cloudflare.com/missing-manuals-io_uring-worker-pool/
	if (unboundedIowMax) {
		if (EMPER_LINUX_LT("5.15")) DIE_MSG("can not set unbounded iow max on linux < 5.15");

		std::array<unsigned int, 2> maxWorkerConfig = {
				0,	// do not change bounded worker count
				unboundedIowMax,
		};

		if (unlikely(io_uring_register_iowq_max_workers(&ring, maxWorkerConfig.data()) < 0))
			DIE_MSG_ERRNO("setting the max unbounded io worker count to " << unboundedIowMax
																																		<< " failed");
	}

	// reserve space for a full SQ
	preparedSqes.reserve(*this->ring.sq.kring_entries);

	if constexpr (emper::IO_COMPLETER_BEHAVIOR != emper::IoCompleterBehavior::none) {
		// In a worker IoContext the eventfd is used to notify the completer thread
		// about completions.
		// In a GlobalIoContext it is used to terminate the completer thread
		notificationEventFd = eventfd(0, 0);
		if (unlikely(notificationEventFd < 0)) {
			DIE_MSG_ERRNO("creating eventfd for io_uring failed");
		}

		// we are in a worker IoContext -> register the eventfd to get notifications
		if (runtime.globalIo) {
			if (unlikely(io_uring_register_eventfd(&ring, notificationEventFd) < 0)) {
				DIE_MSG_ERRNO("io_uring_register_eventfd failed");
			}
			LOGD("registered eventfd " << notificationEventFd << " on io_uring " << ring.ring_fd);
		}
	}

	if constexpr (emper::IO_SINGLE_URING) {
		submitter = new SubmitActor(*this, runtime);
	}
}

IoContext::~IoContext() {
	io_uring_queue_exit(&ring);
	// TODO: check if this is safe
	::close(notificationEventFd);

	delete submitter;
}

auto IoContext::getDefaultUnboundedIowMax() -> unsigned {
	unsigned unboundedIowMax = EMPER_IO_UNBOUNDED_IOW_MAX;

	static const std::string unboundedIowMaxEnvVarName = "EMPER_UNBOUNDED_IOW_MAX";
	auto unboundedIowMaxEnv =
			emper::lib::env::getUnsignedFromEnv<unsigned>(unboundedIowMaxEnvVarName);
	if (unboundedIowMaxEnv) return unboundedIowMaxEnv.value();

	return unboundedIowMax;
}

auto IoContext::getSqHead() const -> unsigned { return *ring.sq.khead; }
auto IoContext::getSqTail() const -> unsigned { return *ring.sq.ktail; }
auto IoContext::getSqEntries() const -> unsigned { return *ring.sq.kring_entries; }
auto IoContext::getSqFlags() const -> unsigned { return *ring.sq.kflags; }

auto IoContext::getCqHead() const -> unsigned { return *ring.cq.khead; }
auto IoContext::getCqHeadSafe() const -> unsigned {
	return reinterpret_cast<std::atomic<unsigned> *>(ring.cq.khead)->load();
}

auto IoContext::getCqTail() const -> unsigned { return *ring.cq.ktail; }
auto IoContext::getCqTailSafe() const -> unsigned {
	return reinterpret_cast<std::atomic<unsigned> *>(ring.cq.ktail)->load();
}

auto IoContext::getCqEntries() const -> unsigned { return *ring.cq.kring_entries; }

auto IoContext::getCqe(unsigned i) const -> struct io_uring_cqe {
	const unsigned mask = *ring.cq.kring_mask;
	const unsigned head = getCqHead();
	return ring.cq.cqes[(head + i) & mask];
}

auto IoContext::getCqFlags() const -> unsigned {
	return *ring.sq.kflags;
}
}	 // namespace emper::io
