// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#pragma once

#include <liburing.h>
#include <liburing/io_uring.h>

#include <array>
#include <atomic>
#include <cassert>
#include <cerrno>
#include <cstddef>
#include <cstdint>
#include <cstring>
#include <functional>	 // for less
#include <iterator>
#include <mutex>
#include <ostream>
#include <string>
#include <utility>
#include <vector>

#include "CallerEnvironment.hpp"	// for CallerEnvironment, EMPER
#include "Common.hpp"
#include "Debug.hpp"	// for LogSubsystem, LogSubsystem::IO, Logger
#include "Emper.hpp"
#include "Fiber.hpp"
#include "Runtime.hpp"	// for Runtime
#include "StealingResult.hpp"
#include "Worker.hpp"
#include "emper-config.h"
#include "io/Future.hpp"
#include "io/Stats.hpp"
#include "io/SubmitActor.hpp"	 // IWYU pragma: keep
#include "lib/TaggedPtr.hpp"
#include "lib/adt/LockedSet.hpp"								 // for LockedSet
#include "sleep_strategy/PipeSleepStrategy.hpp"	 // IWYU pragma: keep

class AbstractWorkStealingScheduler;

namespace emper::io {
class IoContext : public Logger<LogSubsystem::IO> {
	friend class ::Runtime;
	friend class ::AbstractWorkStealingScheduler;
	friend class SubmitActor;
	friend class Future;
	friend class SendFuture;
	friend class RecvFuture;
	friend class CancelWrapper;
	// We need this friend declaration to access protected members through
	// IoContext pointer in GlobalIoContext::globalCompleterFunc
	friend class GlobalIoContext;

	friend class emper::sleep_strategy::PipeSleepStrategy;

	// Debug functions to access the mmaped memory of ring.
	// gdb is not allowed to access the io mmaped memory of the io_uring fd.
	// https://stackoverflow.com/questions/67451177/why-cant-gdb-read-io-uring-cqe-contents
	auto getSqHead() const -> unsigned;
	auto getSqTail() const -> unsigned;
	auto getSqEntries() const -> unsigned;
	auto getSqFlags() const -> unsigned;

	auto getCqHead() const -> unsigned;
	auto getCqHeadSafe() const -> unsigned;
	auto getCqTail() const -> unsigned;
	auto getCqTailSafe() const -> unsigned;
	auto getCqEntries() const -> unsigned;
	auto getCqe(unsigned i) const -> struct io_uring_cqe;
	auto getCqFlags() const -> unsigned;

 protected:
	// Remember the Runtime which created the IoContext
	Runtime &runtime;

	static thread_local IoContext *workerIo;

	// We must synchronize the CQ if it is accessed by multiple threads.
	// This is the case if we use a completer or if other workers try to steal IO.
	static constexpr bool needsCqLock =
			(emper::IO_COMPLETER_BEHAVIOR != emper::IoCompleterBehavior::none) || (emper::IO_STEALING);

	// Are we synchronizing the CQs lockfree
	static constexpr bool locklessCq = needsCqLock && emper::IO_LOCKLESS_CQ;

	// TryLock protecting the completion queue of ring.
	CACHE_LINE_EXCLUSIVE(std::mutex, cq_lock);
	struct io_uring ring;

	// In a worker's IoContext This eventfd is registered with the io_uring to get completion
	// notifications in the globalCompleter.
	// In a GlobalIoContext this eventfd is used to initiate termination
	int notificationEventFd;
	uint64_t notificationEventFdBuf;

	// Remember the worker object for this IoContext
	Worker *worker;

	// SubmitActor used when multiple workers use the SQ of this IoContext.
	// This only happens if emper is build with the IO_SINGLE_URING option
	SubmitActor *submitter = nullptr;

	Stats stats;

	// Members useful for debugging
	std::atomic<long> reqs_in_uring = 0;

	void trackReqsInUring(long requests) {
		if constexpr (emper::DEBUG) {
			reqs_in_uring += requests;
			assert(reqs_in_uring >= 0);
		}
	}

	// We need to keep track of prepared sqes to invalidate them on short submission
	// If a user wants to submit a chain of 5 Futures and the third is invalid
	// io_uring_submit will only submit 3 of 5 prepared sqes leaving the 2 unsubmitted in
	// the SQ.
	// Those unsubmitted sqes would be submitted on the next call of io_uring_submit
	// possibly breaking the guaranties of a chain (cancel all dependent operations).
	// Further more having unsubmitted sqes in the SQ prevents SQ-sized chains from beeing
	// prepared.
	// To prevent all those problem we keep track of the prepared sqes and "invalidate"
	// (prepare as NOP with a nullptr as user_data) submit all unsubmitted  sqes
	// and cancel their futures
	std::vector<struct io_uring_sqe *> preparedSqes;

	// set containing all Futures currently in the io_uring for debugging
	emper::lib::adt::LockedSet<Future *> uringFutureSet;

	/**
	 * @brief recursively prepare a Future and its dependencies
	 *
	 * Futures prepared as dependency will have disabled partial completion
	 * support and are submitted to io_uring with the IOSQE_IO_LINK flag
	 * set.
	 * Abort on preparing a longer list then there a places in the io_uring.
	 *
	 * @param future The future starting the Future chain to prepare
	 * @param chain_length Recursion parameter tracking the resulting chain length
	 *
	 * @return the number of prepared Futures
	 */
	auto prepareFutureChain(Future &future, unsigned chain_length = 1) -> unsigned;

	template <CallerEnvironment callerEnvironment>
	void cancelUnsubmittedChainParts(Future &chain);

	enum class PointerTags : uint16_t { Future, Callback, NewWorkNotification };

	[[nodiscard]] static auto createFutureTag(Future &future) -> emper::lib::TaggedPtr {
		if (future.callback) return {future.callback, static_cast<uint16_t>(PointerTags::Callback)};

		if (!future.isForgotten()) return {&future, static_cast<uint16_t>(PointerTags::Future)};

		return {nullptr};
	}

	/**
	 * @brief submit prepared sqes possibly reaping completions if CQ is full
	 *
	 * @param wait_nr The number of cqes which should be awaited before returning from
	 *        io_uring_submit_and_wait
	 *
	 * @return the number of submitted Futures
	 */
	template <CallerEnvironment callerEnvironment>
	auto submitPreparedSqesAndWait(unsigned wait_nr) -> unsigned;

	/**
	 * @brief submit prepared sqes possibly reaping completions if CQ is full
	 *
	 * @return the number of submitted Futures
	 */
	template <CallerEnvironment callerEnvironment>
	auto submitPreparedSqes() -> unsigned {
		return submitPreparedSqesAndWait<callerEnvironment>(0);
	}

	/**
	 * @brief Submit a future for asynchronous completion in this IoContext
	 *
	 * This function is not part of the public API of the IoContext because a
	 * user code should not be able block the worker thread in the IoContext.
	 *
	 * @param future The Future which should be completed. The Future object
	 *        must stay valid until it is completed.
	 * @param wait_nr The number of cqes which should be awaited before returning from
	 *        io_uring_submit_and_wait
	 */
	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	void submitAndWait(Future &future, unsigned wait_nr);

	inline void setWorkerIo(Worker *worker) {
		// remember our worker
		this->worker = worker;
		workerIo = this;

		if (emper::STATS) {
			stats.workerId = worker->getWorkerId();
		}
	}

	/** The for now relevant data contained in a cqe */
	using Completion = std::pair<int32_t, void *>;

	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	auto getContinuationsFromCompletions(Completion *completions, size_t count, Fiber **continuations)
			-> unsigned {
		unsigned posInBuf = 0;
		for (unsigned i = 0; i < count; ++i) {
			auto &completion = completions[i];
			auto res = completion.first;
			auto *cqe_data = completion.second;

			emper::lib::TaggedPtr tptr(cqe_data);
			// Got a CQE for a forgotten Future.
			if (!tptr) {
				if (res < 0) LOGW("Forgotten future returned " << res << ": " << strerror(-res));
				continue;
			}

			auto tag = static_cast<PointerTags>(tptr.getTag());
			switch (tag) {
				case PointerTags::NewWorkNotification: {
					auto &sleepStrategy =
							reinterpret_cast<PipeSleepStrategy &>(runtime.getWorkerSleepStrategy());
					sleepStrategy.onNewWorkNotification<callerEnvironment>(*this, tptr);
					break;
				}

				case PointerTags::Callback: {
					auto *callback = tptr.getPtr<Future::CallbackInternal>();
					LOGD("Create new callback fiber for " << callback);
					auto *callbackFiber = Fiber::from(
							[&c = *callback, res] {
								c(res);
								delete &c;
							},
							&callback->affinity);

					continuations[posInBuf++] = callbackFiber;
				} break;

				case PointerTags::Future: {
					auto *future = tptr.getPtr<Future>();
					// assert that the future was previously in the uringFutureSet
					assert(uringFutureSet.erase(future) > 0);

					// This is called from all three contexts and must be synchronized
					// using atomics.
					future->recordCompletion(stats, res);

					// It is still safe to call this parallel from all three contexts
					// because possible resubmission is always done in the globalIo
					// when in CallerEnvironment::ANYWHERE and the worker local
					// IoContext otherwise
					// NOTE: This means that IncrementalCompletableFutures now may
					// move between worker IoContexts.
					Fiber *continuation = future->completeAndGetContinuation(res);
					if (continuation) {
						continuations[posInBuf++] = continuation;
					}
				} break;

				default:
					DIE_MSG("Unknown pointer tag encountered: " << (int)tag);
					break;
			}
		}
		return posInBuf;
	}

	template <CallerEnvironment callerEnvironment>
	[[nodiscard]] auto tryReapCompletionWaitFree(Fiber **continuation) -> emper::StealingResult;

	template <CallerEnvironment callerEnvironment, unsigned toReap>
	[[nodiscard]] auto reapCompletionsLockless(Fiber **continuations) -> unsigned;

	template <CallerEnvironment callerEnvironment, unsigned toReap>
	[[nodiscard]] auto reapCompletionsLocked(Fiber **continuations) -> unsigned;

	inline auto getSqe() -> struct io_uring_sqe * {
		struct io_uring_sqe *sqe = io_uring_get_sqe(&ring);
		if constexpr (emper::IO_SQ_POLLER != IoSqPoller::off) {
			// apparently using io_uring_get_sqe after waiting for one getting available
			// with io_uring_sqring_wait may not return a usable sqe.
			// Therefore we loop here
			while (!sqe) {
				LOGD("Poller is behind consuming sqes -> wait");
				int err = io_uring_sqring_wait(&ring);
				if (unlikely(err)) {
					errno = -err;
					DIE_MSG_ERRNO("io_uring_sqring_wait failed");
				}
				LOGD("Done waiting -> get sqe");
				sqe = io_uring_get_sqe(&ring);
			}
		}
		LOGD("Got sqe " << sqe << " from io_uring " << ring.ring_fd);
		assert(sqe);

		return sqe;
	}

 public:
	IoContext(Runtime &runtime, size_t uring_entries, unsigned unboundedIowMax);
	IoContext(Runtime &runtime)
			: IoContext(runtime, EMPER_IO_WORKER_URING_ENTRIES, IoContext::getDefaultUnboundedIowMax()){};
	~IoContext();

	/**
	 * @brief get the default maximum count of unbounded io worker threads
	 *
	 * The default is 0 (unbounded), which can be overwritten during
	 * compile time (-Dio_unbouned_iow_max) or runtime using the
	 * environment variable EMPER_UNBOUNDED_IOW_MAX.
	 *
	 * @return the default maximum count of unbounded io worker threads
	 */
	static auto getDefaultUnboundedIowMax() -> unsigned;

	/**
	 * @brief get IoContext of the current worker
	 *
	 * This function must be called from with in the emper runtime
	 *
	 * @return the thread_local IoContext of the current worker
	 */
	[[nodiscard]] static inline auto getWorkerIo() -> IoContext * {
		emper::assertInRuntime();
		if constexpr (emper::IO_SINGLE_URING) {
			return getIo<CallerEnvironment::ANYWHERE>();
		}

		return workerIo;
	}

	/**
	 * @brief get IoContext for the CallerEnvironment
	 *
	 * This function is for now not independent of a global Runtime reference
	 * because it uses Runtime::getRuntime() if Runtime::getRuntime() would
	 * use a thread_local variable instead of a global one this function
	 * would be like the rest of IoContext independent of a global Runtime
	 * reference
	 *
	 * @return either the worker's IoContext or the GlobalIoContext
	 */
	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	[[nodiscard]] static inline auto getIo() -> IoContext * {
		if constexpr (callerEnvironment == CallerEnvironment::EMPER) {
			return IoContext::getWorkerIo();
		} else {
			// we use a reinterpret_cast here because GlobalIoContext is incomplete
			// at this point but we can't include GlobalIoContext.hpp because it
			// would introduce a cyclic dependency.
			return reinterpret_cast<IoContext *>(Runtime::getRuntime()->globalIo);
		}
	}

	/**
	 * @brief Submit a future for asynchronous completion in this IoContext
	 *
	 * @param future The Future which should be submited to the io_uring.
	 *        The Future object must stay valid until it is completed.
	 */
	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	void submit(Future &future) {
		// Mark the Future as submitted. We do this here because all Futures
		// must go through IoContext::submit even calling Future::submit
		future.markSubmitted();

		// If IO_SYNCHRONOUS is enabled, then all I/O operations will be issued
		// and completed in Future::wait() using a synchronous system call.
		if constexpr (emper::IO_SYNCHRONOUS) return;

		if constexpr (emper::IO_SINGLE_URING) {
			submitter->tell<callerEnvironment>(&future);
		} else {
			submitAndWait(future, 0);
		}
	}

	/**
	 * @brief Submit a future for asynchronous completion in this IoContext
	 *
	 * @param begin Iterator pointing at the start of the Future collection
	 * @param end Iterator pointing at the end of the Future collection
	 *
	 * TODO: gracefully handle cases where there would be more sqes prepared than
	 * there is room in the SQ
	 */
	template <typename InputIt, CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	auto submit(InputIt begin, InputIt end) -> unsigned {
		// Mark Futures as submitted and return if they are completed synchronously
		if constexpr (emper::IO_SYNCHRONOUS) {
			for (InputIt cur = begin; cur != end; ++cur) (*cur)->markSubmitted();
			return std::distance(begin, end);
		}

		for (InputIt cur = begin; cur != end; ++cur) {
			auto *future = *cur;
			future->markSubmitted();

			// This IoContext is Worker exclusive -> prepare the futures
			if constexpr (emper::IO_WORKER_URING) {
				prepareFutureChain(*future);
			}
		}

		// If we have already prepared futures submit them
		if constexpr (emper::IO_WORKER_URING) {
			return submitPreparedSqes<callerEnvironment>();
		}

		// The IoContext is not Worker exclusive ->
		// submit using the IoContexts's SubmitActor
		submitter->tell<callerEnvironment>(begin, end);
		return std::distance(begin, end);
	}

	/**
	 * @brief Submit a future for asynchronous completion in this IoContext and reap completions
	 *
	 * Submit a Future from within the Runtime and reap possible inline completions
	 * directly afterwards
	 *
	 * @param future The Future which should be submited to the io_uring.
	 *        The Future object must stay valid until it is completed.
	 */
	void submitAndReap(Future &future) {
		emper::assertInRuntime();
		submit<CallerEnvironment::EMPER>(future);

		if constexpr (emper::IO_SINGLE_URING || emper::IO_SYNCHRONOUS) {
			return;
		}
		// io_uring will try to synchronously complete any IO request before
		// offloading it to the async backend. See io_uring_enter(2).
		// Immediate offloading can be enforced by setting the IOSQE_ASYNC flag in the sqe.
		// Try to reap a possible synchronous completions.
		reapAndScheduleCompletions<CallerEnvironment::OWNER>();
	}

	/**
	 * @brief return the number of available cqes
	 */
	[[nodiscard]] auto cqeCount() const -> unsigned { return io_uring_cq_ready(&ring); }

	/**
	 * @brief Collect \p toReap fibers waiting on completed IO
	 *
	 * @param[in, out] continuations Buffer big enough to hold \p toReap possible
										 continuation Fibers.
	 *                 Passing the buffer form the caller to the callee allows the buffer to
	 *                 be statically allocated.
	 *
	 * @return The number of continuation Fibers
	 */
	template <CallerEnvironment callerEnvironment, unsigned toReap>
	[[nodiscard]] auto reapCompletions(Fiber **continuations) -> unsigned {
		static_assert(toReap <= CQE_BATCH_COUNT,
									"No need to reap more than there are entries in the CQ");

		// never reap completions on the global IoContext
		assert(this != reinterpret_cast<IoContext *>(runtime.globalIo));

		LOGD("Reaping completions for worker " << std::to_string(worker->getWorkerId()));

		if constexpr (locklessCq) {
			return reapCompletionsLockless<callerEnvironment, toReap>(continuations);
		}

		// CallerEnvironment::EMPER means we are in a stealing worker.
		// Check if the victim's CQ is empty before taking the lock.
		if constexpr (callerEnvironment == CallerEnvironment::EMPER) {
			if (!cqeCount()) return 0;
		}

		return reapCompletionsLocked<callerEnvironment, toReap>(continuations);
	}

	/**
	 * @brief Collect one fibers waiting on completed IO
	 *
	 * @return The fiber or nullptr
	 */
	template <CallerEnvironment callerEnvironment>
	[[nodiscard]] auto reapSingleCompletion() -> Fiber * {
		Fiber *fiber = nullptr;

		if constexpr (emper::WAITFREE_IO_STEALING) {
			auto res = tryReapCompletionWaitFree<callerEnvironment>(&fiber);
			if (res == emper::StealingResult::Stolen) return fiber;
			return nullptr;
		}

		return reapCompletions<callerEnvironment, 1>(&fiber) == 1 ? fiber : nullptr;
	}

	// Should not be more than the uring_entries count.
	static const unsigned CQE_BATCH_COUNT = EMPER_IO_WORKER_URING_ENTRIES;
	using ContinuationBuffer = std::array<Fiber *, CQE_BATCH_COUNT>;

	/**
	 * @brief Collect all fibers waiting on completed IO
	 *
	 * @param[in, out] continuationFibers Buffer big enough to hold all possible continuation Fibers.
	 *                 Passing the buffer form the caller to the callee allows the buffer to
	 *                 be statically allocated.
	 *
	 * @return The number of continuation Fibers
	 */
	template <CallerEnvironment callerEnvironment>
	[[nodiscard]] auto reapCompletions(ContinuationBuffer &continuationFibers) -> unsigned {
		return reapCompletions<callerEnvironment, CQE_BATCH_COUNT>(continuationFibers.data());
	}

	/**
	 * @brief Schedule all fibers waiting on completed IO
	 */
	template <CallerEnvironment callerEnvironment>
	void reapAndScheduleCompletions() {
		ContinuationBuffer completionBuf;
		unsigned ncompletions = reapCompletions<callerEnvironment>(completionBuf);
		if (ncompletions > 0) {
			auto *fibers = completionBuf.data();

			if constexpr (callerEnvironment != CallerEnvironment::ANYWHERE) {
				runtime.schedule(fibers, ncompletions);
			} else {
				runtime.scheduleFromAnywhere(fibers, ncompletions);
			}
		}
	}

	auto getStats() -> Stats & { return stats; };
};
}	 // namespace emper::io
