// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2022 Florian Fischer
#pragma once

#include <atomic>
#include <cstddef>
#include <map>
#include <mutex>
#include <sstream>
#include <string>
#include <utility>
#include <vector>

#include "Common.hpp"

class Runtime;

namespace emper::lib {
/*
 * @brief Runtime local ShardedFileBuffer
 */
class ShardedFileBuffer {
	friend class Runtime;

	static std::atomic<size_t> ids;

	/*
	 * @brief Worker local part of ShardedFileBuffer infrastructure
	 */
	using WorkerBuffers = std::map<size_t, std::stringstream>;

	static thread_local WorkerBuffers* workerBuffers;

	using Shards = std::vector<WorkerBuffers*>;

	static std::mutex registerLock;
	using RuntimeBuffers = std::map<const Runtime*, Shards*>;
	static RuntimeBuffers buffersPerRuntime;

	const Runtime& runtime;
	size_t id;

	std::string file;
	int fd;

 public:
	// TODO: why is Runtime not allowed to use those functions if they are declared private?
	/*
	 * @brief register the current runtime to use ShardedFileBuffers
	 */
	static void registerRuntime(Runtime& runtime);
	/*
	 * @brief Cleanup all runtime specific ShardedFileBuffer data
	 */
	static void unregisterRuntime(const Runtime& runtime);

	ShardedFileBuffer(const Runtime& runtime, int fd);
	ShardedFileBuffer(const Runtime& runtime, const std::string& file);
	~ShardedFileBuffer();

	auto getStream() -> std::stringstream& {
		auto it = workerBuffers->find(id);
		if (it == workerBuffers->end()) {
			auto ret = workerBuffers->emplace(id, std::stringstream());
			if (!ret.second) DIE_MSG("Failed to emplace new stringstream in WorkerBuffers");
			it = ret.first;
		}

		return it->second;
	}

	void put(const std::string& message) { getStream() << message; }

	/*
	 * @brief Collect the buffer shards, join them and writes them to the file
	 *
	 * Attention: It is not safe to call flush during buffer modification!
	 *            Flush the buffer only if you are sure it is not concurrently modified.
	 */
	void flush();
};
}	 // namespace emper::lib
