// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#pragma once

#include <atomic>
#include <cassert>
#include <cstdint>

#include "Common.hpp"
#include "StealingResult.hpp"
#include "emper-common.h"
#include "lib/adt/PushBottomResult.hpp"

namespace adt {

/**
 * @file
 *
 * A CL queue used for work stealing.
 *
 * Implements the partially concurrent queue for the work stealing
 * scheduling algorithm as defined in "Dynamic Circular Work-Stealing Deque" by Chase and Lev
 * [chase2005dynamic]. The worker uses pushBottom to push new items on the stack, and retrieves
 * items via popBottom. Eventually, when its own queue is empty, it will use popTop to retrieve
 * items from a different Worker's queue.
 *
 * Only pushBottom/popBottom and popTop may be used concurrently,
 * i.e. concurrent use of pushBottom with popBottom is not allowed.
 *
 * Unlike the ABP queue, the CL queue uses a circular array and
 * only-incrementing uint64_t bottom and top fields. According to the
 * authors "A 64-bit integer is large enough to accommodate 64 years
 * of pushes, pops and steals executing at a rate of 4 billtion
 * operations per second", so overflows should be no problem.
 */
template <typename PAYLOAD, const uintptr_t CAPACITY>
class WsClv3Queue {
	ALIGN_TO_CACHE_LINE std::atomic<uint64_t> bottom;
	ALIGN_TO_CACHE_LINE std::atomic<uint64_t> top;
	// NOLINTNEXTLINE(modernize-avoid-c-arrays)
	PAYLOAD queue[CAPACITY];

 public:
	// 'bottom' and 'top' are initialized to '1', instead of '0' as
	// it's done in the "Dynamic Circular Work-Stealing Deque" paper
	// because popBottom will decrement bottom, which will result in
	// an underflow if bottom is '0'. The paper's queue uses Java
	// 'long' for bottom and top and is thus safe since it's signed.
	WsClv3Queue() : bottom(1), top(1) {}
	// TODO: Decide what to do regarding the following suppressed lint.
	// NOLINTNEXTLINE(readability-avoid-const-params-in-decls)
	auto pushBottom(const PAYLOAD item) -> PushBottomResult;
	template <const int maxRetries>
	auto popTop(PAYLOAD *item) -> emper::StealingResult;
	auto popBottom(PAYLOAD *item) -> bool;
	[[nodiscard]] auto isFull() const -> bool;
	[[nodiscard]] auto isEmpty() const -> bool;
	void print() const;
	[[nodiscard]] inline auto capacity() const -> uintptr_t { return CAPACITY; }
	[[nodiscard]] inline auto usedSlots() const -> uint64_t { return bottom - top; }
	[[nodiscard]] inline auto freeSlots() const -> uintptr_t { return capacity() - usedSlots(); }
};

template <typename PAYLOAD, const uintptr_t CAPACITY>
auto WsClv3Queue<PAYLOAD, CAPACITY>::pushBottom(const PAYLOAD item) -> PushBottomResult {
	bool pushed = true;
	uint64_t localBottom = bottom.load(std::memory_order_relaxed);
	uint64_t localTop = top.load(std::memory_order_relaxed);

	uint64_t currentSize = localBottom - localTop;
	if (currentSize >= CAPACITY) {
		assert(localBottom == localTop);
		pushed = false;
		goto out;
	}

	queue[bottom % CAPACITY] = item;
	bottom.store(localBottom + 1, std::memory_order_release);

out:
	return PushBottomResult{pushed, pushed ? currentSize + 1 : currentSize};
}

template <typename PAYLOAD, const uintptr_t CAPACITY>
template <const int maxRetries>
auto WsClv3Queue<PAYLOAD, CAPACITY>::popTop(PAYLOAD *item) -> emper::StealingResult {
	int retries = 0;

	uint64_t localTop = top.load(std::memory_order_relaxed);

loop:
	ATTR_UNUSED;	// to suppress unused label 'loop' warning if maxRetries == 0.
	uint64_t localBottom = bottom.load(std::memory_order_acquire);

	if (localBottom <= localTop) return emper::StealingResult::Empty;

	*item = queue[localTop % CAPACITY];

	// If cas fails, then this popTop() lost the race against another
	// popTop().
	if (top.compare_exchange_strong(localTop, localTop + 1, std::memory_order_release,
																	std::memory_order_acquire))
		return emper::StealingResult::Stolen;

	// Loop indefinitely
	if constexpr (maxRetries < 0) goto loop;
	// Loop maxRetries times
	if constexpr (maxRetries > 0) {
		if (retries == maxRetries) return emper::StealingResult::LostRace;

		++retries;
	}

	return emper::StealingResult::LostRace;
}

template <typename PAYLOAD, const uintptr_t CAPACITY>
auto WsClv3Queue<PAYLOAD, CAPACITY>::popBottom(PAYLOAD *item) -> bool {
	uint64_t localBottom = bottom.fetch_sub(1, std::memory_order_acq_rel) - 1;
	uint64_t localTop = top.load(std::memory_order_acquire);

	if (localTop > localBottom) {
		// The queue is empty.
		bottom = localTop;
		return false;
	}

	*item = queue[localBottom % CAPACITY];
	if (localBottom > localTop) return true;

	bool res = top.compare_exchange_weak(localTop, localTop + 1, std::memory_order_release,
																			 std::memory_order_relaxed);

	// Either a popTop() removed the element ('res' is false) or we
	// removed the element ('res' is true), but we need to increment
	// the 'bottom' value, since the element bottom pointed at is now
	// gone. N.B. bottom does point to the next free slot, the actual
	// element we remove is bottom-1.
	bottom.store(localBottom + 1, std::memory_order_relaxed);

	return res;
}

template <typename PAYLOAD, const uintptr_t CAPACITY>
auto WsClv3Queue<PAYLOAD, CAPACITY>::isFull() const -> bool {
	return usedSlots() >= CAPACITY;
}

template <typename PAYLOAD, const uintptr_t CAPACITY>
auto WsClv3Queue<PAYLOAD, CAPACITY>::isEmpty() const -> bool {
	return top >= bottom;
}

}	 // namespace adt
