// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#pragma once

#include <atomic>
#include <chrono>
#include <iostream>
#include <new>
#include <thread>

#include "CallerEnvironment.hpp"
#include "Common.hpp"
#include "Debug.hpp"
#include "Worker.hpp"
#include "emper-common.h"
#include "emper-config.h"
#include "sleep_strategy/AbstractWorkerSleepStrategy.hpp"
#include "sleep_strategy/SleeperState.hpp"
#include "sleep_strategy/Stats.hpp"

#ifdef EMPER_LOCKED_WAKEUP_SEMAPHORE
#include "lib/sync/LockedSemaphore.hpp"
#endif

#ifdef EMPER_POSIX_WAKEUP_SEMAPHORE
#include "lib/sync/PosixSemaphore.hpp"
#endif

#ifdef EMPER_FUTEX_WAKEUP_SEMAPHORE
#include "lib/sync/SpuriousFutexSemaphore.hpp"
#endif

#ifdef EMPER_FUTEX2_WAKEUP_SEMAPHORE
#include "lib/sync/SpuriousFutex2Semaphore.hpp"
#endif

class Runtime;

namespace emper::sleep_strategy {

// Needed to stop waiting for a worker to change its flag if it may be already
// terminated
auto isRuntimeTerminating() -> bool;

template <class Sem>
class AbstractSemaphoreWorkerSleepStrategy
		: AbstractWorkerSleepStrategy<AbstractSemaphoreWorkerSleepStrategy<Sem>>,
			Logger<LogSubsystem::SLEEP_S> {
	const workerid_t workerCount;
	Sem wakeupSem;

	Stats stats;

	// check if the used Semaphore provides a notifySpecific implementation
	static constexpr bool semHasNotifySpecific = requires(Sem s) { s.notifySpecific(0); };

	// check if the used Semaphore needs initialization
	static constexpr bool semNeedsInit = requires(Sem s) { s.init(0); };

	// should the generic notifySpecific implementation be used
	static constexpr bool useGenericNotifySpecificImpl = !semHasNotifySpecific;

	// Member used for the generic notifySpecific implementation
	std::atomic<SleeperState>* states = nullptr;

	// Sleep part of the generic notifySpecific implementation
	inline void genericNotifySpecificSleep() {
		const workerid_t workerId = Worker::getCurrentWorkerId();
		auto& state = states[workerId];

		auto expected = SleeperState::Running;
		if (!state.compare_exchange_strong(expected, SleeperState::Sleeping, std::memory_order_release,
																			 std::memory_order_relaxed)) {
			state.store(SleeperState::Running, std::memory_order_release);
			stats.incSkip();
			LOGD("State " << &state << " was Notified reset to Running and skip sleeping");
			return;
		}

		LOGD("going to sleep");
		wakeupSem.wait();
		stats.incWakeup();

		LOGD("awoken set state to Running");
		state.store(SleeperState::Running, std::memory_order_relaxed);
	}

	// Currently we don't have a good sempahore based algorithm to notify a specific
	// worker.
	// Just brute forcing (notifyAll and hope) is unfortunatly not possbile.
	//
	// A single worker can consume more tokens from the semaphore than others,
	// especially if notifyAll is implemented as `foreach worker: notify()`.
	// This could theoretically lead to a "lost-wakeup" for the targeted worker.
	//
	// 0. Two workers W1, W2
	//    IO completer with wakeup behaviour
	//    inflight IO operation for W1
	// 1. W1 is at the end of its dispatch loop going to sleep
	// 2. W2 is sleeping
	// 3. Kernel produces a cqe in CQ of W1
	// 4. Completer notifies all workers -> sem has value 2
	// 5. W2 iterates the dispatch loop two times without finding work -> goes to sleep
	// 6. W1 calls sleep() with a sem value of 0 -> goes to sleep though it has a cqe in its CQ
	//
	// The used approach is to keep a state for each worker in the states
	// array. Those states make it possible to notify a specific worker.
	// We are sure that the specific worker was successfully notified if we observe
	// that its state has changed.
	// Therefore we can notify all worker in a loop until we observe that the
	// specific worker has changed its state or the runtime is terminating.
	template <CallerEnvironment callerEnvironment>
	inline void genericNotifySpecific(workerid_t workerId) {
		auto& state = states[workerId];
		if (state.exchange(SleeperState::Notified, std::memory_order_release) !=
				SleeperState::Sleeping) {
			LOGD("Worker to notify (" << workerId << ") is not sleeping -> skip notifying");
			return;
		}

		// NotifyAll since we observe that the specific worker has changes its state
		while (state.load(std::memory_order_relaxed) == SleeperState::Notified &&
					 !isRuntimeTerminating()) {
			notifyAll<callerEnvironment>();
			// clang-tidy bug workaround
			// see:
			// https://stackoverflow.com/questions/65564677/clang-tidy-parsing-error-at-spaceship-operator
			std::this_thread::sleep_until(std::chrono::system_clock::now() +
																		std::chrono::milliseconds(1));
			// std::this_thread::sleep_for(std::chrono::milliseconds(1));
		}
	}

	template <CallerEnvironment callerEnvironment>
	[[nodiscard]] inline auto mustNotify() -> bool {
		/* On external work we always increment the semaphore unless we observe
		 * that its value is > workerCount.
		 * If we observe semValue > workerCount we are ensured that some worker will iterate
		 * its dispatchLoop again and must observe the new work.

		 * For work from within emper we could skip wakeup if we observe no one sleeping.
		 * This is sound because wakeupSleepingWorkers() is called from a active
		 * worker which will observe its own new work in its next dispatchLoop before
		 * going to sleep.
		 * BUT this supposed not harmful race may not be harmful for guarantying
		 * progress but it is harmfull for latency.
		 * Inserting new work in the system races with workers about to sleep
		 * actually able to execute the work in a timely manner.

		 * The overhead of increasing the semaphore "to" much are unnessesary atomic
		 * operations on global state. This overhead was not measured in contrast to
		 * the harm caused to latency by beeing greedy when using the semaphore.

		 * Note that sem_getvalue() is allowed to return 0 if there are
		 * waiting workers, hence we need to set the threshold atleast to
		 * 0. This has the disadvantage that we will perform one
		 * unnecessary sem_post. If we are sure the wakeupSem
		 * implementation does not return 0 with waiters,
		 * then the skipWakeupThreshold value could be
		 * reviewed and potentially changed to '-1'.
		 * skipWakeupThreshold = 0;
		 */
		typename Sem::CounterType skipWakeupThreshold;
		if constexpr (callerEnvironment == CallerEnvironment::ANYWHERE) {
			skipWakeupThreshold = workerCount;
		} else {
			skipWakeupThreshold = EMPER_SLEEP_SEM_THRESHOLD;
		}

		auto semValue = wakeupSem.getValue();
		return semValue <= skipWakeupThreshold;
	}

 public:
	AbstractSemaphoreWorkerSleepStrategy(Runtime& runtime, workerid_t workerCount)
			: workerCount(workerCount), stats(runtime) {
		if constexpr (useGenericNotifySpecificImpl) {
			states = new (std::align_val_t(CACHE_LINE_SIZE)) std::atomic<SleeperState>[workerCount];
		}

		if constexpr (semNeedsInit) {
			wakeupSem.init(workerCount);
		}
	}

	~AbstractSemaphoreWorkerSleepStrategy() { delete[] states; }

	[[nodiscard]] inline auto getSleeping() const -> long { return wakeupSem.getValue(); }

	template <CallerEnvironment callerEnvironment>
	inline void notifyOne() {
		if (mustNotify<callerEnvironment>()) {
			LOGD("NotifyOne from " << callerEnvironment);
			wakeupSem.notify();
			stats.incNotify();
		}
	}

	template <CallerEnvironment callerEnvironment>
	inline void notifyMany(unsigned count) {
		if (mustNotify<callerEnvironment>()) {
			LOGD("NotifyMany " << count << " from " << callerEnvironment);
			wakeupSem.notify_many(count);
			stats.incNotify();
			stats.addNotifications(count);
		}
	}

	// notify all workers
	// It is sound to increment the semaphore to much, thus this will only cause
	// workers to iterate the dispatchLoop more often before actually sleeping
	template <CallerEnvironment callerEnvironment>
	inline void notifyAll() {
		if (mustNotify<callerEnvironment>()) {
			LOGD("NotifyAll from " << callerEnvironment);
			wakeupSem.notify_many(workerCount);
			stats.incNotify();
			stats.addNotifications(workerCount);
		}
	}

	template <CallerEnvironment callerEnvironment>
	void notifySpecific(workerid_t workerId) {
		LOGD("specifically notify worker " << workerId << " from " << callerEnvironment);

		if constexpr (semHasNotifySpecific) {
			wakeupSem.notifySpecific(workerId);
		} else {
			genericNotifySpecific<callerEnvironment>(workerId);
		}

		stats.incNotify();
		stats.incNotifications();
	}

	inline void sleep() {
		if constexpr (useGenericNotifySpecificImpl) {
			genericNotifySpecificSleep();
		} else {
			LOGD("going to sleep");
			wakeupSem.wait();
			LOGD("awoken");
			stats.incWakeup();
		}
	}

	void printStats(std::ostream& out = std::cout) { stats.print(out); };
};

#ifdef EMPER_LOCKED_WAKEUP_SEMAPHORE
using SemaphoreWorkerSleepStrategy =
		AbstractSemaphoreWorkerSleepStrategy<::emper::lib::sync::LockedSemaphore>;

#elif defined EMPER_POSIX_WAKEUP_SEMAPHORE
using SemaphoreWorkerSleepStrategy =
		AbstractSemaphoreWorkerSleepStrategy<::emper::lib::sync::PosixSemaphore>;

#elif defined EMPER_FUTEX_WAKEUP_SEMAPHORE
using SemaphoreWorkerSleepStrategy =
		AbstractSemaphoreWorkerSleepStrategy<::emper::lib::sync::SpuriousFutexSemaphore>;

#elif defined EMPER_FUTEX2_WAKEUP_SEMAPHORE
using SemaphoreWorkerSleepStrategy =
		AbstractSemaphoreWorkerSleepStrategy<::emper::lib::sync::SpuriousFutex2Semaphore>;

#else
#error Unknown WorkerSleepSemaphore implementation
#endif

}	 // namespace emper::sleep_strategy
